#include "hardware/gpio.h"
#include "hardware/spi.h"

#include "fonts\6x8_FH.h"
#include "fonts\7x12_FH.h"
#include "fonts\8x16_FH.h"
#include "fonts\30x38_FH.h"

#include "EA_TFT_def.h"

//-------------------------------------------------------------------------------------------------
// global variables

FH_FONT *gFont      = (FH_FONT *)Font_8x16;
uint16_t gColorFG   = WHITE;
uint16_t gColorBG   = BLACK;
int16_t  gZoomX     = 1;
int16_t  gZoomY     = 1;
int16_t  gTFTmodul, gTFTwidth, gTFTheight, gTFTAngle;

//-------------------------------------------------------------------------------------------------

void TFTspi_init(uint clk)
{
  spi_init(TFT_PORT, clk);
  spi_set_format(TFT_PORT, 8, 0,0, SPI_MSB_FIRST);

  gpio_set_function(TFT_SCK, GPIO_FUNC_SPI);
  gpio_set_function(TFT_TX,  GPIO_FUNC_SPI);
  gpio_set_function(TFT_RX,  GPIO_FUNC_SPI);
  
  gpio_set_function(TFT_CS,  GPIO_FUNC_SIO);
  gpio_set_dir(TFT_CS, GPIO_OUT);
  gpio_put(TFT_CS, 1);

  gpio_set_function(TFT_RS,  GPIO_FUNC_SIO);
  gpio_set_dir(TFT_RS, GPIO_OUT);

  gpio_set_function(TFT_RES,  GPIO_FUNC_SIO);
  gpio_set_dir(TFT_RES, GPIO_OUT);
  gpio_put(TFT_RES, 1);
}   

//-------------------------------------------------------------------------------------------------

void write_cmd(uint8_t cmd)
{
  gpio_put(TFT_RS, 0);
  spi_write_blocking(TFT_PORT, &cmd, 1);
}

void write_dat(uint8_t dat)
{
  gpio_put(TFT_RS, 1);
  spi_write_blocking(TFT_PORT, &dat, 1);
}

//-------------------------------------------------------------------------------------------------

void init_ST7735(void)
{
  gpio_put(TFT_CS, 0);

  write_cmd(0x11);	// sleep out
  sleep_ms(120);

  // ST7735S Frame Rate Setting in normal mode: fosc/(((RTNA*2)+40)*(LINE+FPA+BPA+2))=80
  write_cmd(0xB1);	// fosc=850KHz
  write_dat(0x05);	// RTNA=5
  write_dat(0x3C);	// FPA=58
  write_dat(0x3C);	// BPA=58

  // ST7735S Frame Rate Setting in idle mode: fosc/(((RTNB*2)+40)*(LINE+FPB+BPB+2))=80
  write_cmd(0xB2);	// fosc=850KHz
  write_dat(0x05);	//RTNB=5
  write_dat(0x3C);	//FPB=58
  write_dat(0x3C);	//BPB=58

  // ST7735S Frame Rate Setting in parital mode (dot inverson): fosc/(((RTNC*2)+40)*(LINE+FPC+BPC+2))=80
  // ST7735S Frame Rate Setting in parital mode (column inverson): fosc/(((RTNC*2)+40)*(LINE+FPC+BPC+2))=80
  write_cmd(0xB3);	// fosc=850KHz
  write_dat(0x05);	// RTNC=5
  write_dat(0x3C);	// FPC=58
  write_dat(0x3C);	// BPC=58
  write_dat(0x05);	// RTND=5
  write_dat(0x3C);	// FPD=58
  write_dat(0x3C);	// BPD=58

  // ST7735S Display Inversion Control
  write_cmd(0xB4);	// Dot inversion: 20184019 modify from Sitronix initial code
  write_dat(0x07);	// 0xB4[2]=Inversion setting in normal mode
                  	// 0xB4[1]=Inversion setting in idle mode
                  	// 0xB4[0]=Inversion setting in partial mode

  // ST7735S Power on Sequence
  write_cmd(0xC0);	// power control 1
  write_dat(0xE9);	// {Par.3[0],Par.1[4:0]}=VRHP[5:0]=2->GVDD=4.6, Par.1[7:5]=AVDD [2:0]=6->AVDD=5.1
  write_dat(0x09);	// {Par.3[1],Par.2[4:0]}=VRHN[5:0]=2->GVCL=-4.6
  write_dat(0x04);	// Par.3[7:6]=MODE[1:0]=2x

  write_cmd(0xC1);	// power control 2
  write_dat(0xC5);	// Par.1 [1:0]=VGHBT[1:0]=0->VGH=2*AVDD+VGH25-0.5
                  	// Par.1 [3:2]=VGLSEL[1:0]=0->VGL=-7.5
                  	// Par.1 [7:6]=VGLSEL[1:0]=3->VGH25=2.4

  write_cmd(0xC2);	// power control 3
  write_dat(0x0D);	// (Sitronix initial) {Par.1 [7:6],Par.2[7:0]=DCA[9:0]=000000000'b->Booster set up cycle BCLK/1 BCLK/3 BCLK/1 BCLK/1 BCLK/1 in normal mode
  write_dat(0x00);	// Par.1[5:3]=SAPA[2:0]=001'b-> OP current is small in normal mode
                  	// Par.1[2:0]=APA[2:0]=001'b-> OP current is Large in normal mode

  write_cmd(0xC3);      //power control 4
  write_dat(0x8D);	// (Sitronix initial){Par.1 [7:6],Par.2[7:0]=DCB[9:0]=1001101010'b->Booster set up cycle BCLK/2 BCLK/1 BCLK/2 BCLK/2 BCLK/2 in idle mode
  write_dat(0x6A);	// Par.1[5:3]=SAPB[2:0]=001'b-> OP current is small in idle mode
                  	// Par.1[2:0]=APB[2:0]=011'b-> OP current is Medium in idle mode

  write_cmd(0xC4);	// power control 5
  write_dat(0x8D);	// (Sitronix initial){Par.1 [7:6],Par.2[7:0]=DCC[9:0]=1011101110'b->Booster set up cycle BCLK/2 BCLK/2 BCLK/2 BCLK/4 BCLK/2 in partial mode
  write_dat(0xEE);	// Par.1[5:3]=SAPC[2:0]=001'b-> OP current is small in partial mode

  // ST7735S VCOM
  write_cmd(0xC5);	// VCOM setting value
  write_dat(0x15);	// 0xC5[5:0]=010010'b -> VCOM=-0.875

  write_cmd(0x21);	// ST7735 Display Inversion on

  // ST7735 Gamma Sequence
  write_cmd(0xE0);	// Gamma setting value (Positive Polarity)
  write_dat(0x07);	// Par.1[5:0]=VRF0P[5:0]=000011'b (Variable Resistor VRHP)
  write_dat(0x0E);	// Par.2[5:0]=VOS0P[5:0]=011011'b (Variable Resistor VRLP)
  write_dat(0x08);	// Par.3[5:0]=PK0P[5:0]=010010'b  (Voltage of V3 grayscale)
  write_dat(0x07);	// Par.4[5:0]=PK1P[5:0]=010001'b  (Voltage of V4 grayscale)
  write_dat(0x10);	// Par.5[5:0]=PK2P[5:0]=111111'b  (Voltage of V12 grayscale)
  write_dat(0x07);	// Par.6[5:0]=PK3P[5:0]=111010'b  (Voltage of V20 grayscale)
  write_dat(0x02);	// Par.7[5:0]=PK4P[5:0]=111010'b  (Voltage of V28 grayscale)
  write_dat(0x07);	// Par.8[5:0]=PK5P[5:0]=110100'b  (Voltage of V36 grayscale)
  write_dat(0x09);	// Par.9[5:0]=PK6P[5:0]=101111'b  (Voltage of V44 grayscale)
  write_dat(0x0F);	// Par.10[5:0]=PK7P[5:0]=101011'b  (Voltage of V52 grayscale)
  write_dat(0x25);	// Par.11[5:0]=PK8P[5:0]=110000'b  (Voltage of V56 grayscale)
  write_dat(0x36);	// Par.12[5:0]=PK9P[5:0]=111010'b  (Voltage of V60 grayscale)
  write_dat(0x00);	// Par.13[5:0]=SELV0P[5:0]=000000'b  (Voltage of V0 grayscale)
  write_dat(0x08);	// Par.14[5:0]=SELV1P[5:0]=000001'b  (Voltage of V1 grayscale)
  write_dat(0x04);	// Par.15[5:0]=SELV62P[5:0]=000010'b  (Voltage of V62 grayscale)
  write_dat(0x10);	// Par.16[5:0]=SELV63P[5:0]=001001'b  (Voltage of V63 grayscale)

  write_cmd(0xE1);	// Gamma setting value (Negative Polarity)
  write_dat(0x0A);	// Par.1[5:0]=VRF0N[5:0]=000011'b (Variable Resistor VRHN)
  write_dat(0x0D);	// Par.2[5:0]=VOS0N[5:0]=011011'b (Variable Resistor VRLN)
  write_dat(0x08);	// Par.3[5:0]=PK0N[5:0]=010010'b  (Voltage of V3 grayscale)
  write_dat(0x07);	// Par.4[5:0]=PK1N[5:0]=010001'b  (Voltage of V4 grayscale)
  write_dat(0x0F);	// Par.5[5:0]=PK2N[5:0]=110010'b  (Voltage of V12 grayscale)
  write_dat(0x07);	// Par.6[5:0]=PK3N[5:0]=101111'b  (Voltage of V20 grayscale)
  write_dat(0x02);	// Par.7[5:0]=PK4N[5:0]=101010'b  (Voltage of V28 grayscale)
  write_dat(0x07);	// Par.8[5:0]=PK5N[5:0]=101111'b  (Voltage of V36 grayscale)
  write_dat(0x09);	// Par.9[5:0]=PK6N[5:0]=101110'b  (Voltage of V44 grayscale)
  write_dat(0x0F);	// Par.10[5:0]=PK7N[5:0]=101100'b  (Voltage of V52 grayscale)
  write_dat(0x25);	// Par.11[5:0]=PK8N[5:0]=111001'b  (Voltage of V56 grayscale)
  write_dat(0x35);	// Par.12[5:0]=PK9N[5:0]=111111'b  (Voltage of V60 grayscale)
  write_dat(0x00);	// Par.13[5:0]=SELV0N[5:0]=000000'b  (Voltage of V0 grayscale)
  write_dat(0x09);	// Par.14[5:0]=SELV1N[5:0]=000000'b  (Voltage of V1 grayscale)
  write_dat(0x04);	// Par.15[5:0]=SELV62N[5:0]=000001'b  (Voltage of V62 grayscale)
  write_dat(0x10);	// Par.16[5:0]=SELV63N[5:0]=001001'b  (Voltage of V63 grayscale)

  write_cmd(0xFC);	// Enable Gate power save mode
  write_dat(0xC0);	// 0xFC[7:6]=GCV_Enable[1:0]=10'b-> Gate Pump Clock Frequency disable
                  	// 0xFC[3:2]=CLK_Variable[1:0]=11'b-> Save Power Ability is Large

  write_cmd(0x3A);	// Color Mode
  write_dat(0x05);	// 65K Mode RGB565

  write_cmd(0x29);	// Display on

  gpio_put(TFT_CS, 1);
}

void init_ST7789(void)
{
  gpio_put(TFT_CS, 0);

  write_cmd(0x11);  // Sleep out
  sleep_ms(10);

//--------------------------------ST7789V Frame rate setting------------------------------------//
  write_cmd(0xb2); 
  write_dat(0x05); 
  write_dat(0x05); 
  write_dat(0x00);
  write_dat(0x33);
  write_dat(0x33);

  write_cmd(0xb7);
  write_dat(0x35);

//---------------------------------ST7789V Power setting----------------------------------------//
  write_cmd(0xb8);
  write_dat(0x2f);
  write_dat(0x2b);
  write_dat(0x2f);

  write_cmd(0xbb);
  write_dat(0x2B);

  write_cmd(0xc0);
  write_dat(0x2c);

  write_cmd(0xc2);
  write_dat(0x01);

  write_cmd(0xc3);
  write_dat(0x0b);

  write_cmd(0xc4);
  write_dat(0x20);

  write_cmd(0xc6);
  write_dat(0x11);

  write_cmd(0xd0);
  write_dat(0xa4);
  write_dat(0xa1);

  write_cmd(0xe8);
  write_dat(0x03);

  write_cmd(0xe9);
  write_dat(0x0d);
  write_dat(0x12);
  write_dat(0x00);

//-------------------------------ST7789V gamma setting-------------------------------------------//
  write_cmd(0xe0);
  write_dat(0xd0);
  write_dat(0x06);
  write_dat(0x0b);
  write_dat(0x0a);
  write_dat(0x09);
  write_dat(0x05);
  write_dat(0x2e);
  write_dat(0x43);
  write_dat(0x44);
  write_dat(0x09);
  write_dat(0x16);
  write_dat(0x15);
  write_dat(0x23);
  write_dat(0x27);

  write_cmd(0xe1);
  write_dat(0xd0); 
  write_dat(0x06);
  write_dat(0x0b); 
  write_dat(0x09);
  write_dat(0x08);
  write_dat(0x06);
  write_dat(0x2e);
  write_dat(0x44);
  write_dat(0x44);
  write_dat(0x3a); 
  write_dat(0x15);
  write_dat(0x15);
  write_dat(0x23);
  write_dat(0x26);

//------------------------------display and color format setting--------------------------------//
  write_cmd(0x3A); // Set color mode
  write_dat(0x05); // RGB 65K Colors, Control interface 16bit/pixel

//--------------Display on-------------------
  write_cmd(0x21);
  write_cmd(0x13);
  write_cmd(0x29);

  gpio_put(TFT_CS, 1);
}

void init_HX8357(void)
{
  gpio_put(TFT_CS, 0);

  write_cmd(0x11); //Sleep Out
  sleep_ms(150);

  write_cmd(0xB1); //SETPower
  write_dat(0x00); //STB
  write_dat(0x16); //
  write_dat(0x1C); //VSPR = 4.41V
  write_dat(0x1C); //VSNR = -4.41V
  write_dat(0xC3); //AP 0xc3
  write_dat(0x5C); //FS 0x44
  
  write_cmd(0xB3);
  write_dat(0x80);  //Set SDO_EN
  write_dat(0x00);
  write_dat(0x06);
  write_dat(0x06);

  write_cmd(0x3A); //COLMOD
  write_dat(0x05); //RGB565 16-Bit
  
  write_cmd(0x21);
  write_cmd(0x13); //Normal display mode
  write_cmd(0x29); //Display On
  
  gpio_put(TFT_CS, 1);
}


void TFTinit(int16_t tfttyp)
{
  gTFTmodul = tfttyp;

  switch(gTFTmodul)
   { case EA_TFT009:  gTFTwidth =  80;  gTFTheight = 160;  init_ST7735();  break;
     case EA_TFT015:  gTFTwidth = 240;  gTFTheight = 240;  init_ST7789();  break;
     case EA_TFT020:  gTFTwidth = 240;  gTFTheight = 320;  init_ST7789();  break;
     case EA_TFT028:  gTFTwidth = 240;  gTFTheight = 320;  init_ST7789();  break;
     case EA_TFT035:  gTFTwidth = 320;  gTFTheight = 480;  init_HX8357();  break;
   }

  gTFTAngle = 0;
}

void TFTreset(int16_t tfttyp)
{
  gpio_put(TFT_RES, 0);
  sleep_us(20);
  gpio_put(TFT_RES, 1);
  sleep_ms(120);

  TFTinit(tfttyp);
}

//-------------------------------------------------------------------------------------------------

void TFTwindow(int16_t xs, int16_t ys, int16_t xe, int16_t ye)
{
  write_cmd(0x36);	  // Memory data access control
  switch(gTFTmodul)
   { case EA_TFT009:  switch(gTFTAngle)
                       { case 0:    write_dat(0xC8);  xs += 26;  xe += 26;  ys +=  1;  ye +=  1;  break;
                         case 90:   write_dat(0x68);  xs +=  1;  xe +=  1;  ys += 26;  ye += 26;  break;
                         case 180:  write_dat(0x08);  xs += 26;  xe += 26;  ys +=  1;  ye +=  1;  break;
                         case 270:  write_dat(0xA8);  xs +=  1;  xe +=  1;  ys += 26;  ye += 26;  break;
                       }
                   break;

     case EA_TFT015:  switch(gTFTAngle)
                       { case 0:    write_dat(0x00);  break;
                         case 90:   write_dat(0xA0);  xs += 80;  xe += 80;  break;
                         case 180:  write_dat(0xC0);  ys += 80;  ye += 80;  break;
                         case 270:  write_dat(0x60);  break;
                       }
                   break;
     
     case EA_TFT020:
     case EA_TFT028:  switch(gTFTAngle)
                       { case 0:    write_dat(0x00);  break;
                         case 90:   write_dat(0xA0);  break;
                         case 180:  write_dat(0xC0);  break;
                         case 270:  write_dat(0x60);  break;
                       }
                   break;

     case EA_TFT035:  switch(gTFTAngle)
                       { case 0:    write_dat(0x48);  break;
                         case 90:   write_dat(0xE8);  break;
                         case 180:  write_dat(0x88);  break;
                         case 270:  write_dat(0x28);  break;
                       }
                   break;
   }

  write_cmd(0x2A);      // Column address set
  write_dat(HI(xs));
  write_dat(LO(xs));
  write_dat(HI(xe));
  write_dat(LO(xe));

  write_cmd(0x2B);      // Row address set
  write_dat(HI(ys));
  write_dat(LO(ys));
  write_dat(HI(ye));
  write_dat(LO(ye));

  write_cmd(0x2C);    	// Memory write

  gpio_put(TFT_RS, 1);  // put RS to HI for following data
}

//-------------------------------------------------------------------------------------------------

void TFTfill(uint16_t color, int16_t xs, int16_t ys, int16_t xe, int16_t ye)
{
  int16_t anzX = xe-xs+1;
  int16_t anzY = ye-ys+1;
  uint8_t *pbuf, linebuf[anzX*2];

  gpio_put(TFT_CS, 0);
  TFTwindow(xs,ys, xe,ye);

  pbuf = linebuf;
  while(xs++ <= xe)
   { *pbuf++ = HI(color);   // write color to linebuf
     *pbuf++ = LO(color);
   }
  while(anzY--)
     spi_write_blocking(TFT_PORT, linebuf, anzX*2);

  gpio_put(TFT_CS, 1);
}

void TFTclear(uint16_t color)
{
  TFTfill(color, 0, 0, gTFTwidth-1, gTFTheight-1);
}

//-------------------------------------------------------------------------------------------------

void zoombyte(uint8_t *pd, uint8_t b, int16_t zoom)
{
  switch(zoom)
   { case 1:  pd[0] = b;
           break;

     case 2:  pd[0] = pd[1] = 0;
              if(b & 0x80) pd[0]  = 0b11000000;
              if(b & 0x40) pd[0] |= 0b00110000;
              if(b & 0x20) pd[0] |= 0b00001100;
              if(b & 0x10) pd[0] |= 0b00000011;
              if(b & 0x08) pd[1]  = 0b11000000;
              if(b & 0x04) pd[1] |= 0b00110000;
              if(b & 0x02) pd[1] |= 0b00001100;
              if(b & 0x01) pd[1] |= 0b00000011;
           break;

     case 3:  pd[0] = pd[1] = pd[2] = 0;
              if(b & 0x80)   pd[0]  = 0b11100000;
              if(b & 0x40)   pd[0] |= 0b00011100;
              if(b & 0x20) { pd[0] |= 0b00000011;
                             pd[1]  = 0b10000000; }
              if(b & 0x10)   pd[1] |= 0b01110000;
              if(b & 0x08)   pd[1] |= 0b00001110;
              if(b & 0x04) { pd[1] |= 0b00000001;
                             pd[2]  = 0b11000000; }
              if(b & 0x02)   pd[2] |= 0b00111000;
              if(b & 0x01)   pd[2] |= 0b00000111;
           break;

     case 4:  pd[0] = pd[1] = pd[2] = pd[3] = 0;
              if(b & 0x80) pd[0]  = 0b11110000;
              if(b & 0x40) pd[0] |= 0b00001111;
              if(b & 0x20) pd[1]  = 0b11110000;
              if(b & 0x10) pd[1] |= 0b00001111;
              if(b & 0x08) pd[2]  = 0b11110000;
              if(b & 0x04) pd[2] |= 0b00001111;
              if(b & 0x02) pd[3]  = 0b11110000;
              if(b & 0x01) pd[3] |= 0b00001111;
           break;
   }
}

void getlinedat(uint8_t *pd, uint8_t *ps, int16_t byteanz, int16_t zoom)
{
  while(byteanz--)
   { zoombyte(pd, *ps++, zoom);
     pd += zoom;
   }
}

void TFTtext(int16_t xs, int16_t ys, char *ptext)
{
  if(gFont != NULL)
   { uint16_t color;
     int16_t  x,y, xe,ye, z;
     int16_t  width  = gFont->width  * gZoomX;
     int16_t  height = gFont->height * gZoomY;
     uint8_t *pdat, *pline, charline[gFont->bytewidth*gZoomX], *pbuf, linebuf[width*2];
     uint8_t  b, msk;

     ye = ys + height - 1;
     
     while(*ptext)
      { if(*ptext >= gFont->first  &&  *ptext <= gFont->last)
         { pdat = gFont->dat + (int)gFont->bytechar * (*ptext - gFont->first);
           xe = xs + width  - 1;

           gpio_put(TFT_CS, 0);
           TFTwindow(xs,ys, xe,ye);
           for(y=0; y < gFont->height; y++)
            { getlinedat(charline, pdat, gFont->bytewidth, gZoomX);
              pdat += gFont->bytewidth;
              
              pbuf  = linebuf;
              pline = charline;
              msk   = 0x80;
              for(x=0; x < width; x++)
               { if(msk == 0x80)
                    b = *pline++;
                 color = (b & msk) ? gColorFG : gColorBG;
                 *pbuf++ = HI(color); *pbuf++ = LO(color);   // write color to linebuf
                 msk = (msk == 0x01) ? 0x80 : msk >> 1;
               }
              for(z=0; z < gZoomY; z++)
                 spi_write_blocking(TFT_PORT, linebuf, width*2);
            }
           gpio_put(TFT_CS, 1);
         }
        ptext++;
        xs += width;
      }
   }
}

//-------------------------------------------------------------------------------------------------

void TFTbitmap(int16_t xs, int16_t ys, const unsigned char *pbitmap)
{
  BMP_HDR *pBMP = (BMP_HDR *)pbitmap;
 
  if(pBMP->bfType == 0x4D42 )	// always BM, no RLE compressed BMPs
   { if(((pBMP->biBitCount <= 8 || pBMP->biBitCount == 24)  &&  pBMP->biCompression == 0)  ||  (pBMP->biBitCount == 16  &&  pBMP->biCompression == 3))
      { uint16_t w, palette[256];
        uint8_t *pdat, *pbuf, linebuf[pBMP->biWidth*2];
        int16_t  x, width,height, byteperline;
        uint8_t  r,g,b, msk;
        
        width  = pBMP->biWidth;
        height = pBMP->biHeight;
        byteperline = (width * pBMP->biBitCount + 31) / 32 * 4;

        if(pBMP->biBitCount <= 8)
         { if((w=pBMP->biClrUsed) == 0)
            { if(pBMP->biBitCount == 1)  w = 2;
              if(pBMP->biBitCount == 4)  w = 16;
              if(pBMP->biBitCount == 8)  w = 256;
            }
           pdat = (uint8_t *)pbitmap + pBMP->bfOffBits - w*4;
           for(x=0; x < w; x++)
            { palette[x] = RGB565(pdat[2], pdat[1], pdat[0]);
              pdat += 4;
            }
         }

        gpio_put(TFT_CS, 0);
        TFTwindow(xs,ys, xs+width-1, ys+height-1);
        do
         { pbuf = linebuf;
           pdat = (uint8_t *)pbitmap + pBMP->bfOffBits + byteperline * --height;

           switch(pBMP->biBitCount)
            {
              case 1:  msk = 0x80;
                       for(x=0; x < width; x++)
                        { if(msk == 0x80)
                             b = *pdat++;
                          w = (b & msk) ? palette[1] : palette[0];
                          *pbuf++ = HI(w); *pbuf++ = LO(w);   // write color to linebuf
                          msk = (msk == 0x01) ? 0x80 : msk >> 1;
                        }
                     break;

              case 4:  for(x=0; x < width; x++)
                        { if(x & 1)
                             w = palette[b & 0x0F];
                          else
                           { b = *pdat++;
                             w = palette[b >> 4];
                           }
                          *pbuf++ = HI(w); *pbuf++ = LO(w);   // write color to linebuf
                        }
                     break;

              case 8:  for(x=0; x < width; x++)
                       { w = palette[*pdat++];
                         *pbuf++ = HI(w); *pbuf++ = LO(w);   // write color to linebuf
                       }
                     break;

              case 16: for(x=0; x < width; x++)
                        { w  = *pdat++;
                          w += *pdat++ << 8;
                         *pbuf++ = HI(w); *pbuf++ = LO(w);   // write color to linebuf
                        }
                     break;

              case 24: for(x=0; x < width; x++)
                        { b = *pdat++;
                          g = *pdat++;
                          r = *pdat++;
                          w = RGB565(r,g,b);
                         *pbuf++ = HI(w); *pbuf++ = LO(w);   // write color to linebuf
                        }
                     break;
            }
           spi_write_blocking(TFT_PORT, linebuf, width*2);
         }
        while(height > 0);

        gpio_put(TFT_CS, 1);
      }
   }
}

//----------------------------------------------------------------------------
//                            XX  
// +----+                   +----+     
// |    |      +------+     |    |     +------+
// | 0° |    XX| 90°  |     |180°|     | 270° |XX
// +----+      +------+     +----+     +------+
//   XX 
//----------------------------------------------------------------------------

void setTFTangle(int16_t angle)
{
  if(angle == 0  ||  angle == 180  ||  angle == 90  ||  angle == 270)
   {
     if( ((angle== 0 || angle==180) && (gTFTAngle==90 || gTFTAngle==270)) ||
         ((angle==90 || angle==270) && (gTFTAngle== 0 || gTFTAngle==180))   )
      { int16_t tmp = gTFTwidth;  // swap gTFTwidth <-> gTFTheight
         gTFTwidth  = gTFTheight;
         gTFTheight = tmp;
      }

     gTFTAngle = angle;
   }
}

void setTFTfont(int16_t nr, int16_t zx, int16_t zy)
{
  FH_FONT *pf = NULL;

  switch(nr)
   { case FONT_6X8:  	pf = (FH_FONT *)Font_6x8;	  break;
     case FONT_7X12:  pf = (FH_FONT *)Font_7x12;  break;
     case FONT_8X16:  pf = (FH_FONT *)Font_8x16;  break;
     case FONT_30X38: pf = (FH_FONT *)Font_30x38;	break;
   }

  if(pf != NULL  &&  pf->F == 'F'  &&  pf->H == 'H')
     gFont = pf;

  if(zx >= 1  &&  zx <= 4)
     gZoomX = zx;

  if(zy >= 1  &&  zy <= 4)
     gZoomY = zy;
}

void setTFTcolor(uint16_t fg, uint16_t bg)
{
  gColorFG = fg;
  gColorBG = bg;
}
